/**
 ** Copyright (c) 2002 Sun Microsystems, Inc.
 **
 ** See the file "license.terms" for information on usage and redistribution
 ** of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 **/

/**
 **  gcspy_space.cpp
 **
 **  Represents a space
 **/

#include <stdio.h>

#include "gcspy_space.hpp"

gcspySpace::gcspySpace (const char       *name,
			const char       *driver_name,
			const char       *title,
			const char       *block_info,
			gcspy_length_t    tile_num,
			const char       *unused_string,
			gcspy_bool_t      main_space) {
  _changed = 0;

  gcspyUtils::copy_string(_name, name, name_len);
  gcspyUtils::copy_string(_driver_name, driver_name, name_len);
  _tile_num = 0;
  _tile_names = NULL;
  if (unused_string != NULL)
    gcspyUtils::copy_string(_unused_string, unused_string, unused_string_len);
  else
    gcspyUtils::copy_string(_unused_string, default_unused_string, unused_string_len);
  gcspyUtils::copy_string(_title, title, name_len);
  gcspyUtils::copy_string(_block_info, block_info, block_info_len);
  _main_space = main_space;
  _control = NULL;
  set_space_info("");
  
  _stream_num = 0;
  for (int i = 0; i < max_streams; ++i)
    _streams[i] = NULL;

  setup_arrays(tile_num);
}

void
gcspySpace::setup_arrays (gcspy_length_t new_tile_num) {
  if (_tile_names != NULL) {
    gcspy_assert(_tile_num > 0);
    for (int i = 0; i < _tile_num; ++i)
      if (_tile_names[i] != NULL)
	delete _tile_names[i];
    delete _tile_names;
    _tile_names = NULL;
  } else {
    gcspy_assert(_tile_num == 0);
  }
  if (_control != NULL) {
    gcspy_assert(_tile_num > 0);
    delete _control;
    _control = NULL;
  } else {
    gcspy_assert(_tile_num == 0);
  }

  if (new_tile_num > 0) {
    _tile_names = new char*[new_tile_num];
    if (_tile_names == NULL) {
      char buffer[128];
      sprintf(buffer,
	      "Space[%d]: new failed for tile name array of len %u",
	      _id, new_tile_num);
      gcspy_raise_error(buffer);
    }
    for (int i = 0; i < new_tile_num; ++i)
      _tile_names[i] = NULL;

    _control = new char[new_tile_num];
    if (_control == NULL) {
      char buffer[128];
      sprintf(buffer,
	      "Space[%d]: new failed for control array of len %u",
	      _id, new_tile_num);
      gcspy_raise_error(buffer);
    }
  } else {
    _tile_names = NULL;
    _control = NULL;
  }

  _tile_num = new_tile_num;
}

void
gcspySpace::add_stream (gcspyStream *stream) {
  if (_stream_num == max_streams) {
    char buffer[128];
    sprintf(buffer, "Space[%d]: max stream num reached (%d)",_id, max_streams);
    gcspy_raise_error(buffer);
  }

  stream->set_id(_stream_num);
  stream->setup_arrays(_tile_num);
  _streams[_stream_num] = stream;
  ++_stream_num;
}

void
gcspySpace::reset_data () {
  for (int i = 0; i < _stream_num; ++i) {
    _streams[i]->reset_data(_tile_num);
    _streams[i]->reset_summary();
  }
  reset_control();
}

void
gcspySpace::reset_control () {
  gcspy_assert((_tile_num > 0) || (_control == NULL));
  for (int i = 0; i < _tile_num; ++i)
    _control[i] = control_used;
}

void
gcspySpace::serialise (gcspyClient *client) {
  gcspyBufferedOutput output = gcspyBufferedOutput(client);
  serialise(&output);
  output.close();
  output.send();
}

/**
 **    Driver info format when serialised
 **    ----------------------------------
 **      id               short
 **      name             string
 **      driver_name      string
 **      title            string
 **      block_info       string
 **      tile_num         int
 **      unused_string    string
 **      main_space       boolean
 **      stream_num       short
 **      ( stream_info )*
 **      ( tileName )*    string
 **    ----------------------------------
 **/

void
gcspySpace::serialise (gcspyBufferedOutput *output) {
  int i;
  output->write_short((short) _id);
  output->write_string(_name);
  output->write_string(_driver_name);
  output->write_string(_title);
  output->write_string(_block_info);
  output->write_int(_tile_num);
  output->write_string(_unused_string);
  output->write_boolean(_main_space);
  output->write_short((short) _stream_num);
  for (i = 0; i < _stream_num; ++i)
    _streams[i]->serialise(output);
  for (i = 0; i < _tile_num; ++i)
    output->write_string(get_tile_name(i));
}

void
gcspySpace::resize (gcspy_length_t new_tile_num) {
  for (int i = 0; i < _stream_num; ++i)
    _streams[i]->setup_arrays(new_tile_num);
  setup_arrays(new_tile_num);
  _changed = 1;
}
