/**
 ** Copyright (c) 2002 Sun Microsystems, Inc.
 **
 ** See the file "license.terms" for information on usage and redistribution
 ** of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 **/

/**
 **  gcspy_stream.cpp
 **
 **  Represents a stream
 **/

#include <stdio.h>

#include "gcspy_utils.hpp"
#include "gcspy_stream.hpp"

gcspyStream::gcspyStream (unsigned          data_type,
			  const char       *name,
			  int               min_value,
			  int               max_value,
			  int               zero_value,
			  int               default_value,
			  const char       *string_pre,
			  const char       *string_post,
			  unsigned          presentation,
			  unsigned          paint_style,
			  unsigned          max_stream_index,
			  gcspyColor       *color,
			  gcspy_bool_t      enable_summary) {
  _data_type = data_type;
  gcspyUtils::copy_string(_name, name, string_len);
  _min_value = min_value;
  _max_value = max_value;
  _zero_value = zero_value;
  _default_value = default_value;
  gcspyUtils::copy_string(_string_pre, string_pre, string_len);
  gcspyUtils::copy_string(_string_post, string_post, string_len);
  _presentation = presentation;
  _paint_style = paint_style;
  _max_stream_index = max_stream_index;
  _color = *color;
  _enum_num = 0;
  _enum_names = NULL;
  _data = NULL;
  _summary = NULL;
  _enum_names = NULL;
  if (enable_summary)
    setup_summary();

  if (_presentation == presentation_enum) {
    gcspy_assert(_enum_names == NULL);
    gcspy_assert(_enum_num == 0);
    _enum_names = new char*[enum_max_num];
    if (_enum_names == NULL) {
      char buffer[128];
      sprintf(buffer, "new failed for the enum array of len %u", enum_max_num);
      gcspy_raise_error(buffer);
    }
    for (int i = 0; i < enum_max_num; ++i)
      _enum_names[i] = NULL;
  }
}

void
gcspyStream::add_enum_name (gcspy_index_t index,
			    const char *name) {
  gcspy_assert(index < enum_max_num);
  gcspy_assert(_enum_names != NULL);
  gcspy_assert(_enum_names[index] == NULL);
  gcspy_assert(index == _enum_num);

  _enum_names[index] = gcspyUtils::clone_string(name);
  if (_enum_names[index] == NULL) {
      char buffer[128];
      sprintf(buffer, "new failed for an enum name");
      gcspy_raise_error(buffer);
  }
  ++_enum_num;
}

void
gcspyStream::setup_summary () {
  gcspy_assert(_summary == NULL);

  gcspy_length_t length;
  switch (_presentation) {
  case presentation_plain:
  case presentation_plus:
  case presentation_max_var:
    length = 1;
    break;
  case presentation_percent:
  case presentation_percent_var:
    length = 2;
    break;
  case presentation_enum:
    length = enum_max_num;
    break;
  }
  _summary = new int[length];
  if (_summary == NULL) {
    char buffer[128];
    sprintf(buffer,
	    "Stream[%d]: new failed for summary array of len %u",
	    _id, length);
    gcspy_raise_error(buffer);
  }
}

void
gcspyStream::setup_arrays (gcspy_length_t tile_num) {
  if (_data != NULL)
    delete _data;

  if (tile_num > 0) {
    switch (_data_type) {
    case byte_type:
      _data = new char[tile_num];
      break;
    case short_type:
      _data = (char *) new short[tile_num];
      break;
    case int_type:
      _data = (char *) new int[tile_num];
      break;
    default:
      gcspy_assert(0);
      break;
    }
    if (_data == NULL) {
      char buffer[128];
      sprintf(buffer,
	      "Stream[%d]: new failed for data array of len %u",
	      _id, tile_num);
      gcspy_raise_error(buffer);
    }
  } else {
    _data = NULL;
  }
}

void
gcspyStream::reset_data (gcspy_length_t tile_num) {
  gcspy_assert((tile_num > 0) || (_data == NULL));

  switch (_data_type) {
  case byte_type: {
    char *curr = _data;
    char *limit = curr + tile_num;
    while (curr < limit) {
      *curr = (char) _default_value;
      ++curr;
    }
    break;
  }
  case short_type: {
    short *curr = (short *) _data;
    short *limit = curr + tile_num;
    while (curr < limit) {
      *curr = (short) _default_value;
      ++curr;
    }
    break;
  }
  case int_type: {
    int *curr = (int *) _data;
    int *limit = curr + tile_num;
    while (curr < limit) {
      *curr = _default_value;
      ++curr;
    }
    break;
  }
  default:
    gcspy_assert(0);
    break;
  }
}

void
gcspyStream::reset_summary () {
  if (_summary != NULL) {
    gcspy_length_t len = get_summary_len();
    for (int i = 0; i < len; ++i)
      _summary[i] = 0;
  }
}

/**
 **    Stream info format when serialised
 **    ----------------------------------
 **      id                        byte
 **      data_type                 byte
 **      name                      string
 **      min_value                 int
 **      max_value                 int
 **      zero_value                int
 **      default_value             int
 **      string_pre                string
 **      string_post               string
 **      presentation              byte
 **      paint_style               byte
 **      max_stream_index          byte
 **      color                     color
 **      enum_name_num             byte
 **    ----------------------------------
 **      ( enum names )*           string
 **/

void
gcspyStream::serialise (gcspyBufferedOutput *output) {
  output->write_byte(_id);
  output->write_byte(_data_type);
  output->write_string(_name);
  output->write_int(_min_value);
  output->write_int(_max_value);
  output->write_int(_zero_value);
  output->write_int(_default_value);
  output->write_string(_string_pre);
  output->write_string(_string_post);
  output->write_byte((char) _presentation);
  output->write_byte((char) _paint_style);
  output->write_byte((char) _max_stream_index);
  output->write_color(&_color);
  output->write_byte((char) _enum_num);
  gcspy_assert( (_enum_num == 0) || (_enum_names != NULL) );
  for (int i = 0; i < _enum_num; ++i) {
    gcspy_assert( _enum_names[i] != NULL );
    output->write_string(_enum_names[i]);
  }
}

  void gcspyStream::increment (gcspy_index_t index, int value) {
	  gcspy_assert(index >= 0 && index <= _max_stream_index);  //TODO: check >= <=
	  switch (_data_type) {
		  case (byte_type):
			  get_byte_data()[index] += (char) value;
			  break;
		  case (short_type):
			  get_short_data()[index] += (short) value;
			  break;
		  case (int_type):
			  get_int_data()[index] += (int) value;
			  break;
	  }
  }

  void gcspyStream::distribute(int start, int remainder, int blockSize, int value) {
	  gcspy_assert(remainder <= blockSize);
	    if (value <= remainder) {  // fits in this tile
	    	increment(start, value);
	    } else {
	    	increment(start, remainder);
	      value -= remainder;
	      start++;
	      while (value >= blockSize) {
	    	  increment(start, blockSize);
	        value -= blockSize;
	        start++;
	      }
	      increment(start,value);
	    }

}
