/**
 ** $Id: Space.java 28 2005-06-20 13:13:35Z rej $
 ** Copyright (c) 2002 Sun Microsystems, Inc.
 **
 ** See the file "license.terms" for information on usage and redistribution
 ** of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 **/

package gcspy.interpreter;

import gcspy.comm.BufferedInput;
import gcspy.comm.BufferedOutput;

/**
 * Represents a space
 * 
 * @author Tony Printezis
 */
public class Space {

	// Constants
	/** Control value for a used tile */
	static public final byte CONTROL_USED = 1;
	/** Control value for a background tile */
	static public final byte CONTROL_BACKGROUND = 2;
	/** Control value for an unused tile */
	static public final byte CONTROL_UNUSED = 4;
	/**
	 * Control value for a separator. The separator goes <emph>before</emph> the
	 * tile
	 */
	static public final byte CONTROL_SEPARATOR = 8;
	/**
	 * Control value for a control link. The link goes <emph>after</emph> the
	 * tile
	 */
	static public final byte CONTROL_LINK = 16;

	/** Default string for unused tiles */
	static protected final String DEFAULT_UNUSED_STRING = "NOT USED";

	/** Default string for tiles with an illegal value */
	static protected final String DEFAULT_ILLEGAL_STRING = "Illegal value";

	/**
	 * Is this value a background control?
	 * 
	 * @param val
	 *            A control
	 * @return true if the tile is a background tile
	 */
	static public boolean isControlBackground(byte val) {
		return (val & CONTROL_BACKGROUND) != 0;
	}
	/**
	 * Is this value a link control?
	 * 
	 * @param val
	 *            A control
	 * @return true if the tile is a control link
	 */
	static public boolean isControlLink(byte val) {
		return (val & CONTROL_LINK) != 0;
	}
	/**
	 * Is this value a separator control?
	 * 
	 * @param val
	 *            A control
	 * @return true if the tile is a separator
	 */
	static public boolean isControlSeparator(byte val) {
		return (val & CONTROL_SEPARATOR) != 0;
	}
	/**
	 * Is this value an unused control?
	 * 
	 * @param val
	 *            A control
	 * @return true if the tile is unused
	 */
	static public boolean isControlUnused(byte val) {
		return (val & CONTROL_UNUSED) != 0;
	}
	/**
	 * Is this value a used control?
	 * 
	 * @param val
	 *            A control
	 * @return true if the tile is used
	 */
	static public boolean isControlUsed(byte val) {
		return (val & CONTROL_USED) != 0;
	}
	// Configuration
	/** The space's ID */
	protected int id;
	/** The space's name */
	protected String name;
	/** The name of the space's driver */
	protected String driverName;
	/** The space's title */
	protected String title;
	/** The block information for a space */
	protected String blockInfo;
	/** The number of tiles */
	protected int tileNum;
	/** The names of the tiles */
	protected String tileNames[];
	/** The number of streams */
	protected int streamNum;

	/******************** Data Setting ********************/

	/** The streams */
	protected Stream streams[];

	/** The space information */
	protected String spaceInfo;

	/** The string to use for unused tiles */
	protected String unusedString;
	
	/** Is this the main space? */
	protected boolean mainSpace;


	/******************** Accessor Methods ********************/

	/** The controls for this space */
	protected byte control[];

	/******************** Constructors ********************/

	private Space() {
	}

	/**
	 * Create a new space preinitialized with some values
	 * 
	 * @param name
	 *            The space's name
	 * @param driverName
	 *            Its driver's name
	 * @param tileNum
	 *            The number of tiles in the space
	 * @param title
	 *            The title to use
	 * @param blockInfo
	 *            The block information tile to use
	 * @param streamNum
	 *            The number of streams in the space
	 * @param unusedString
	 *            The string to use for unused tiles
	 * @param mainSpace
	 *            Is this the main space?
	 */
	public Space(String name, String driverName, int tileNum,
			String title, String blockInfo, int streamNum, String unusedString,
			boolean mainSpace) {
		this.name = name;
		this.driverName = driverName;
		this.tileNum = tileNum;
		this.title = title;
		this.blockInfo = blockInfo;
		this.tileNames = null;
		this.streams = new Stream[streamNum];
		this.streamNum = 0;
		if (unusedString != null)
			this.unusedString = unusedString;
		else
			this.unusedString = DEFAULT_UNUSED_STRING;
		this.mainSpace = mainSpace;

		setupControl();
	}

	/**
	 * Add a stream to this space
	 * 
	 * @param stream
	 *            The stream to add
	 * @return An ID for this stream
	 */
	public int addStream(Stream stream) {
		int streamID = streamNum;
		stream.setID(streamID);
		streams[streamID] = stream;

		++streamNum;

		return streamID;
	}

	/**
	 * Deserialise the space
	 * 
	 * @param input
	 *            The BufferedInput to use
	 */
	public static Space deserialise(BufferedInput input) {
		Space result = new Space();
		result.id = (int) input.readShort();
		result.name = input.readString();
		result.driverName = input.readString();
		result.title = input.readString();
		result.blockInfo = input.readString();
		result.tileNum = input.readInt();
		result.unusedString = input.readString();
		result.mainSpace = input.readBoolean();
		result.streamNum = (int) input.readShort();
		result.streams = new Stream[result.streamNum];
		for (int i = 0; i < result.streamNum; ++i) {
			result.streams[i] = Stream.deserialise(input);
			result.streams[i].setSpace(result);
		}
		result.deserialiseTileNames(input);
		result.setupControl();
		return result;
	}

	private void deserialiseTileNames(BufferedInput input) {
		tileNames = new String[tileNum];
		for (int i = 0; i < tileNum; ++i)
			tileNames[i] = input.readString();
	}

	/******************** Debugging ********************/

	public void dump() {
		System.out.println("== Space[" + id + "]: " + getFullName());
		System.out.println("--   " + tileNum + " tiles");
		System.out.println("--   " + streamNum + " streams");
		System.out.println("--   tile names");
		for (int i = 0; i < tileNum; ++i)
			System.out.println("--     " + tileNames[i]);
		for (int i = 0; i < streamNum; ++i)
			streams[i].dump();
	}

	/**
	 * Get the space's block information
	 * 
	 * @return the block information for this space
	 */
	public String getBlockInfo() {
		return blockInfo;
	}
	

	/**
	 * Get the controls for this space
	 * 
	 * @return the controls
	 */
	public byte[] getControl() {
		return control;
	}	

	/**
	 * Get the space's full name
	 * 
	 * @return the name of the space and its driver
	 */
	public String getFullName() {
		return name + "  [" + driverName + "]";
	}

	/**
	 * Get the space's ID
	 * 
	 * @return the space ID
	 */
	public int getID() {
		return id;
	}

	/**
	 * Get the space's name
	 * 
	 * @return the space's name
	 */
	public String getName() {
		return name;
	}

	/**
	 * Get the space information for this space
	 * 
	 * @return the space information
	 */
	public String getSpaceInfo() {
		return spaceInfo;
	}

	/** ****************** Presentation ******************* */
	
	/**
	 * Represent a summary of the space as a string
	 * 
	 * @param prefix1
	 *            A string to prefix the summary
	 * @param prefix2
	 *            A string to prefix the summary if the style is
	 *            <code>PRESENTATION_ENUM</code>
	 * @return a string representation of the summary
	 */
	public String getSpaceString(String prefix1, String prefix2) {
		String str = "";
		if (spaceInfo != null)
			str += spaceInfo;
		for (int i = 0; i < streamNum; ++i) {
			Stream stream = streams[i];
			str += stream.presentFullSummary(prefix1, prefix2);
		}
		return str;
	}

	/******************** Utilities ********************/

	/**
	 * Return a stream
	 * 
	 * @param streamID
	 *            the ID of the stream
	 * @return the stream
	 */
	public Stream getStream(int streamID) {
		return streams[streamID];
	}

	/**
	 * Get the number of streams in this space
	 * 
	 * @return the number of streams
	 */
	public int getStreamNum() {
		return streamNum;
	}
	
	/**
	 * Return the name of a tile
	 * 
	 * @param i
	 *            the index of the tile
	 * @return its name
	 */
	public String getTileName(int i) {
		return tileNames[i];
	}

	/**
	 * Get the number of tiles in this space
	 * 
	 * @return the number of tiles
	 */
	public int getTileNum() {
		return tileNum;
	}	

	/******************** Control ********************/

	/**
	 * Return a string representation of the tile with the given index
	 * 
	 * @param index
	 *            the tile index
	 * @return its string representation
	 */
	public String getTileString(int index) {
		StringBuilder str = new StringBuilder();
		str.append(title);
		str.append(index);
		if (!isControlBackground(control[index])) {
			String tileName = tileNames[index];
			str.append(tileName);
			str.append('\n');
			str.append(blockInfo);
		} else {
			str.append('\n');
		}
		if (isControlUsed(control[index])) {
			for (int i = 0; i < streamNum; ++i) {
				Stream stream = streams[i];
				if (stream.getAccessor().isDataAvailable()) {
					str.append(stream.presentData(index));
					str.append('\n');
				}
			}
		} else if (isControlBackground(control[index])) {
			str.append("BACKGROUND");
		} else if (isControlUnused(control[index])) {
			str.append(unusedString);
		}
		return str.toString();
	}

	/**
	 * Get the space's title
	 * 
	 * @return the title
	 */
	public String getTitle() {
		return title;
	}

	/**
	 * Get the string used for unused tiles
	 * 
	 * @return the string for unused tiles
	 */
	public String getUnusedString() {
		return unusedString;
	}

	/**
	 * Returns the string for tiles with illegal values.
	 */
	public String getIllegalString() {
		// Could be specified by server in the future
		return DEFAULT_ILLEGAL_STRING;
	}

	/** Initialise all the controls to unused */
	public void initControl() {
		for (int i = 0; i < control.length; ++i)
			control[i] = CONTROL_USED;
	}

	/**
	 * Is this the main space in the visualiser?
	 * 
	 * @return true if this is the main space
	 */
	public boolean isMainSpace() {
		return mainSpace;
	}

	/******************** Serialisation / Deserialisation ********************/

	/**
	 * Resize this space
	 * 
	 * @param tileNum
	 *            The number of tiles in the resized space
	 */
	public void resize(int tileNum) {
		this.tileNum = tileNum;
		for (int i = 0; i < streamNum; ++i)
			streams[i].initData(tileNum);
		setupControl();
	}

	/**
	 * Serialise the space
	 * 
	 * @param output
	 *            The BufferedOutput to use
	 */
	public void serialise(BufferedOutput output) {
		output.writeShort((short) id);
		output.writeString(name);
		output.writeString(driverName);
		output.writeString(title);
		output.writeString(blockInfo);
		output.writeInt(tileNum);
		output.writeString(unusedString);
		output.writeBoolean(mainSpace);
		output.writeShort((short) streamNum);
		for (int i = 0; i < streamNum; ++i)
			streams[i].serialise(output);
		serialiseTileNames(output);
	}

	private void serialiseTileNames(BufferedOutput output) {
		for (int i = 0; i < tileNum; ++i)
			output.writeString(tileNames[i]);
	}

	/**
	 * Set the controls
	 * 
	 * @param control
	 *            the controls
	 */
	public void setControl(byte control[]) {
		this.control = control;
	}
	
	/**
	 * Set the control for a tile
	 * 
	 * @param tag
	 *            The control value
	 * @param index
	 *            The tile index
	 */
	public void setControl(byte tag, int index) {
		if (isControlBackground(tag) || isControlUnused(tag)) {
			if (isControlUsed(control[index]))
				control[index] &= (~CONTROL_USED);
		}

		control[index] |= tag;
	}
	

	/**
	 * Set the control values for a range of tiles
	 * 
	 * @param tag
	 *            The control value
	 * @param start
	 *            The starting tile
	 * @param len
	 *            The number of tiles
	 */
	public void setControlRange(byte tag, int start, int len) {
		for (int i = start; i < (start + len); ++i) {
			setControl(tag, i);
		}
	}

	/**
	 * Set this space's ID
	 * 
	 * @param id
	 *            the ID
	 */
	public void setID(int id) {
		this.id = id;
	}

	/**
	 * Set the space information
	 * 
	 * @param spaceInfo
	 *            The space information
	 */
	public void setSpaceInfo(String spaceInfo) {
		this.spaceInfo = spaceInfo;
	}

	public void setTileName(int tile, String name) {
		tileNames[tile] = name;
	}

	/** Setup the controls and initialise them to unused */
	protected void setupControl() {
		control = new byte[tileNum];
		initControl();
	}

}
