/**
 ** $Id: ServerInterpreter.java 29 2005-06-20 14:03:41Z rej $
 ** Copyright (c) 2002 Sun Microsystems, Inc.
 **
 ** See the file "license.terms" for information on usage and redistribution
 ** of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 **/

package gcspy.interpreter.server;

import gcspy.comm.BufferedInput;
import gcspy.comm.BufferedOutput;
import gcspy.comm.Command;
import gcspy.comm.SocketClient;
import gcspy.comm.SocketServer;
import gcspy.interpreter.EventFilters;
import gcspy.interpreter.Events;
import gcspy.interpreter.Interpreter;
import gcspy.interpreter.InterpreterException;
import gcspy.interpreter.OutputGenerator;
import gcspy.interpreter.Space;
import gcspy.interpreter.Stream;
import gcspy.interpreter.StreamData;
import gcspy.interpreter.TraceData;
import gcspy.utils.Timer;
import gcspy.utils.Utils;

import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * The server interpreter
 * 
 * @author Tony Printezis
 */
public class ServerInterpreter extends Interpreter implements Runnable {
	
	private static final Logger logger = Logger.getLogger(ServerInterpreter.class.getName());
	
	private SocketServer server;
	private int maxLen;

	private boolean connected;

	private boolean paused;
	private boolean pauseNow;

	private boolean playOne;

	private boolean shutdownReq;
	private boolean shutdownSent;

	// of true, server counts events and times
	private boolean collectStats;

	// counters
	private Timer elapsedTimer;
	private Timer compensationTimer;
	private boolean singleServe;

	/** ****************** Accessor Methods ******************* */

	/**
	 * Get a space
	 * 
	 * @param i
	 *            The space's ID
	 * @return The space
	 */
	private Space getSpace(int i) {
		return getData().getSpace(i);
	}

	/**
	 * Add a server space to the interpreter
	 * 
	 * @param space
	 *            The server space
	 * @return an ID for the space
	 */
	public int addSpace(Space space) {
		return getData().addSpace(space);
	}

	/**
	 * Transmit all the streams for all the spaces known to this interpreter.
	 * 
	 * @throws IOException
	 */
	public void transmitStreams() throws IOException {
		for (int i = 0; i < getData().getSpaceNum(); ++i) {
			Space space = getSpace(i);
			sendSpace(space);
			for (int j = 0; j < space.getStreamNum(); ++j) {
				Stream stream = space.getStream(j);
				StreamData data = stream.getData();
				if (data == null)
					continue; // no data for this stream

				sendStream(i, j, data);
				sendSummary(i, j, stream.getSummary());
			}
			sendSpaceInfo(i, space.getSpaceInfo());
			sendControl(i, space.getControl());
		}
	}

	/** ****************** Commands ******************* */

	/** *** PAUSE REQ **** */

	private class PauseReqCmd implements Command {
		public void execute(BufferedInput input) {
			pauseNow = true;
		}
	}

	/** *** PAUSE **** */

	private class PauseCmd implements Command {
		public void execute(BufferedInput input) {
			throw new InterpreterException("PAUSE Command not implemented");
		}
	}

	/**
	 * Send a PAUSE command
	 * 
	 * @throws IOException
	 */
	public void sendPause() throws IOException {
		sendSingleCommand(PAUSE_CMD);
	}

	/** *** RESTART **** */

	private class RestartCmd implements Command {
		public void execute(BufferedInput input) {
			paused = false;
		}
	}

	/** *** PLAY ONE **** */

	private class PlayOneCmd implements Command {
		public void execute(BufferedInput input) {
			playOne = true;
		}
	}

	/** *** SHUTDOWN REQ **** */

	private class ShutdownReqCmd implements Command {
		public void execute(BufferedInput input) {
			shutdownReq = true;
		}
	}

	/** *** SHUTDOWN **** */

	private class ShutdownCmd implements Command {
		public void execute(BufferedInput input) {
			throw new InterpreterException("SHUTDOWN Command not implemented");
		}
	}

	/**
	 * Send a SHUT_DOWN command
	 * 
	 * @throws IOException
	 */
	public void sendShutdown() throws IOException {
		sendSingleCommand(SHUTDOWN_CMD);
	}

	/** *** STREAM **** */

	private class StreamCmd implements Command {
		public void execute(BufferedInput input) {
			throw new InterpreterException("STREAM Command not implemented");
		}
	}

	/**
	 * Send a stream for a space
	 * 
	 * @param spaceID
	 *            The space's ID
	 * @param streamID
	 *            The stream's ID
	 * @param data
	 *            The stream's data
	 * @throws IOException
	 */
	public void sendStream(int spaceID, int streamID, StreamData data)
			throws IOException {
		sendStream(spaceID, streamID, data, data.size());
	}

	/**
	 * Send a stream for a space
	 * 
	 * @param spaceID
	 *            The space's ID
	 * @param streamID
	 *            The stream's ID
	 * @param data
	 *            The stream's data
	 * @param len
	 *            The length of the stream data
	 * @throws IOException
	 */
	public void sendStream(int spaceID, int streamID, StreamData data, int len) throws IOException {
		BufferedOutput output = client.createBufferedOutput();

		start(output);
		putCmd(output, STREAM_CMD);
		output.writeByte((byte) spaceID);
		output.writeByte((byte) streamID);
		switch (data.getType()) {
		case Byte:
			output.writeByteArray(data.getBytes(), len);
			break;
		case Short:
			output.writeShortArray(data.getShorts(), len);
			break;
		case Int:
			output.writeIntArray(data.getInts(), len);
			break;
		}
		finish(output);

		client.send(output);
	}

	/** *** EVENT **** */

	private class EventCmd implements Command {
		public void execute(BufferedInput input) {
			throw new InterpreterException("EVENT Command not implemented");
		}
	}

	/**
	 * Send an event
	 * 
	 * @param eventID
	 *            The event's ID
	 * @param elapsedTime
	 *            The elapsed time for the event
	 * @param compensationTime
	 *            The compensation time for the event
	 * @throws IOException
	 */
	public void sendEvent(int eventID, int elapsedTime, int compensationTime)
			throws IOException {
		BufferedOutput output = client.createBufferedOutput();

		start(output);
		putCmd(output, EVENT_CMD);
		output.writeByte((byte) eventID);
		output.writeInt(elapsedTime);
		output.writeInt(compensationTime);
		finish(output);

		client.send(output);
	}

	/** *** CONTROL **** */

	private class ControlCmd implements Command {
		public void execute(BufferedInput input) {
			throw new InterpreterException("CONTROL Command not implemented");
		}
	}

	/**
	 * Send a control stream for a space
	 * 
	 * @param spaceID
	 *            The space's ID
	 * @param control
	 *            The controls
	 * @throws IOException
	 */
	public void sendControl(int spaceID, byte control[]) throws IOException {
		BufferedOutput output = client.createBufferedOutput();

		start(output);
		putCmd(output, CONTROL_CMD);
		output.writeByte((byte) spaceID);
		output.writeByteArray(control);
		finish(output);

		client.send(output);
	}

	/** *** EVENT FILTERS **** */

	private class EventFiltersCmd implements Command {
		public void execute(BufferedInput input) {
			// compiler is being dense
			EventFilters eventFilters = getData().getEventFilters();

			int len = (int) input.readShort();
			boolean enabled[] = new boolean[len];
			int delays[] = new int[len];
			boolean pauses[] = new boolean[len];
			int periods[] = new int[len];

			for (int i = 0; i < len; ++i) {
				enabled[i] = input.readBoolean();
				delays[i] = input.readInt();
				pauses[i] = input.readBoolean();
				periods[i] = input.readInt();
			}

			eventFilters.setEnabled(enabled);
			eventFilters.setDelays(delays);
			eventFilters.setPauses(pauses);
			eventFilters.setPeriods(periods);
		}
	}

	/** *** EVENT COUNT **** */

	private class EventCountCmd implements Command {
		public void execute(BufferedInput input) {
			throw new InterpreterException(
					"EVENT COUNT Command not implemented");
		}
	}

	/**
	 * Send the event count
	 * 
	 * @throws IOException
	 */
	public void sendEventCount() throws IOException {
		BufferedOutput output = client.createBufferedOutput();

		start(output);
		putCmd(output, EVENT_COUNT_CMD);
		output.writeIntArray(getData().getEventCount());
		finish(output);

		client.send(output);
	}

	/** *** SUMMARY **** */

	private class SummaryCmd implements Command {
		public void execute(BufferedInput input) {
			throw new InterpreterException("SUMMARY Command not implemented");
		}
	}

	/**
	 * Send the summary
	 * 
	 * @param spaceID
	 *            The space's ID
	 * @param streamID
	 *            The stream's ID
	 * @param summary
	 *            The summary values
	 * @throws IOException
	 */
	public void sendSummary(int spaceID, int streamID, int summary[])
			throws IOException {
		BufferedOutput output = client.createBufferedOutput();

		start(output);
		putCmd(output, SUMMARY_CMD);
		output.writeByte((byte) spaceID);
		output.writeByte((byte) streamID);
		if (summary == null)
			output.writeEmptyArray();
		else
			output.writeIntArray(summary);
		finish(output);

		client.send(output);
	}

	/** *** SPACE INFO **** */

	private class SpaceInfoCmd implements Command {
		public void execute(BufferedInput input) {
			throw new InterpreterException("SPACE INFO Command not implemented");
		}
	}

	/**
	 * Send the space information
	 * 
	 * @param spaceID
	 *            The space's ID
	 * @param spaceInfo
	 *            The space information
	 * @throws IOException
	 */
	public void sendSpaceInfo(int spaceID, String spaceInfo) throws IOException {
		BufferedOutput output = client.createBufferedOutput();

		start(output);
		putCmd(output, SPACE_INFO_CMD);
		output.writeByte((byte) spaceID);
		output.writeString(spaceInfo);
		finish(output);

		client.send(output);
	}

	/** *** SPACE **** */

	/*
	private class SpaceCmd implements Command {
		public void execute(BufferedInput input) {
			throw new InterpreterException("SPACE Command not implemented");
		}
	}
	*/

	/**
	 * Send a space
	 * 
	 * @param space
	 *            The space
	 * @throws IOException
	 */
	public void sendSpace(Space space) throws IOException {
		BufferedOutput output = client.createBufferedOutput();

		start(output);
		putCmd(output, SPACE_CMD);
		space.serialise(output);
		finish(output);
		client.send(output);
	}

	/** ****************** Configuration Communication ******************* */

	private class ClientOutputGenerator implements OutputGenerator {

		private SocketClient client;

		public BufferedOutput createBufferedOutput() throws IOException {
			return client.createBufferedOutput();
		}

		public void writeBuffer(int len) throws IOException {
			client.send(len);
		}

		ClientOutputGenerator(SocketClient client) {
			this.client = client;
		}
	}

	private void serialiseEverything() throws IOException {
		getData().serialiseEverything(new ClientOutputGenerator(client));
	}

	/** ****************** Connection ******************* */

	private boolean isConnected() {
		return connected;
	}

	private void checkClient() throws IOException {
		BootstrapParameters params = new BootstrapParameters();

		/* client sends first */
		receiveBootInfo(true, params);
		sendBootInfo(true, null);

		if (params.pauseAtStart) {
			pauseNow = true;
		}
	}

	private void waitForNewClient(int maxLen) throws IOException {
		logger.log(Level.INFO, "waiting for client to connect");
		/*
		 * make sure we clear up the client variable before we re-assign it,
		 * otherwise we need double the memory requirements
		 */
		client = null;
		client = server.waitForNewClient(maxLen);
		logger.log(Level.INFO, "client connected");
	}

	private void waitUntilClientConnected() {
		while (!isConnected()) {
			Utils.sleep(10);
		}

		Utils.sleep(50);
		safepoint();
	}
	
	/**
	 * Start the server
	 * 
	 * @param port
	 *            The port on which to communicate
	 * @throws IOException
	 */
	public void startServer(int port) throws IOException {
		startServer(port, false);
	}

	/**
	 * Start the server
	 * 
	 * @param port
	 *            The port on which to communicate
	 * @param wait
	 *            Whether to wait until client connects
	 * @throws IOException
	 */
	public void startServer(int port, boolean wait) throws IOException {
		startServer(port, wait, DEFAULT_MAX_LEN);
	}

	/**
	 * Start the server
	 * 
	 * @param port
	 *            The port on which to communicate
	 * @param wait
	 *            Whether to wait until client connects
	 * @param maxLen
	 *            The buffer length
	 * @throws IOException
	 */
	public void startServer(int port, boolean wait, int maxLen)
			throws IOException {
		logger.log(Level.INFO, "starting server, port " + port);
		server = new SocketServer(port);
		logger.log(Level.FINE, "started server, port " + getBindPort());

		if (wait)
			logger.log(Level.FINE, "blocked until client connects");

		this.maxLen = maxLen;
		new Thread(this).start();

		if (wait)
			waitUntilClientConnected();
	}
	
	public void setSingleServe() {
		this.singleServe = true;
	}
	
	public int getBindPort() {
		if (server == null) return SocketServer.ANY_PORT;
		return server.getBindPort();
	}

	/** ****************** Timers ******************* */

	/** Start the compensation timer */
	public void startCompensationTimer() {
		compensationTimer.start();
	}

	/** Stop the compensation timer */
	public void stopCompensationTimer() {
		compensationTimer.stop();
	}

	/** Reset the compensation and elapsed timers */
	private void resetTimers() {
		compensationTimer.reset();
		elapsedTimer.reset();
		elapsedTimer.start();
	}

	/** ****************** Main Loop ******************* */

	/** Start the main interpreter thread loop */
	public void run() {
		mainThreadLoop();
	}

	private void connectToClient() throws IOException {
		waitForNewClient(maxLen);
		internalSetup();
		logger.log(Level.FINE, "checking client");
		checkClient();
		logger.log(Level.FINE, "sending boot info");
		serialiseEverything();

		if (collectStats)
			resetTimers();

		connected = true;
	}

	private void mainLoop() throws IOException {
		logger.log(Level.FINE, "starting server main loop");
		while (true) {
			client.receive();
			if (client.hasTerminated()) {
				logger.log(Level.WARNING, "main loop terminated (client died)");
				connected = false;
				break;
			}

			BufferedInput input = client.createBufferedInput();
			execute(input);
			if (shutdownReq) {
				logger.log(Level.INFO, "main loop terminated (shutdown request received)");
				break;
			}
		}
	}

	private void mainThreadLoop() {
		try {
			do {
				connectToClient();
				mainLoop();
				
				if (shutdownReq) {
					while (!shutdownSent) {
						Utils.sleep(10);
					}
					connected = false;
				}

				client.close();
			} while (!this.singleServe);
		} catch (IOException e) {
			logger.log(Level.SEVERE, "Exception caught, terminating", e);
			System.exit(-1);
		}
	}

	/** ****************** Safepoints ******************* */

	private void internalSafepoint() {
		try {
			if (shutdownReq) {
				sendShutdown();
				shutdownSent = true;
			}

			if (pauseNow) {
				logger.log(Level.FINE, "paused");
				sendPause();
				pauseNow = false;
				paused = true;
			}

			if (paused) {
				while (paused && isConnected()) {
					Utils.sleep(10);
					if (shutdownReq) {
						sendShutdown();
						shutdownSent = true;
						paused = false;
					}
					if (playOne) {
						logger.log(Level.FINE, "play one");
						playOne = false;
						break;
					}
				}
				if ((!paused) && (!shutdownReq))
					logger.log(Level.FINE, "restarted");
			}
		} catch (IOException e) {
			logger.log(Level.SEVERE, "Exception caught, terminating", e);
			System.exit(-1);
		}
	}

	/** Indicate a safe point */
	public void safepoint() {
		if (isConnected()) {
			internalSafepoint();
		}
	}

	/**
	 * Update the event counter for an event
	 * 
	 * @param eventID
	 *            The event's ID
	 */
	public void updateEventCounter(int eventID) {
		++getData().getEventCount()[eventID];
	}

	private void internalEventBoundary(int eventID, int elapsedTime,
			int compensationTime) {
		try {
			transmitStreams();
			sendEventCount();
			sendEvent(eventID, elapsedTime, compensationTime);

			int delay = getData().getEventFilters().getDelays()[eventID];
			if (delay > 0) {
				Utils.sleep(delay);
			}

			if (getData().getEventFilters().getPauses()[eventID]) {
				pauseNow = true;
			}

			internalSafepoint();
		} catch (IOException e) {
			logger.log(Level.SEVERE, "Exception caught, terminating", e);
			System.exit(-1);
		}
	}

	/**
	 * Indicate an event boundary
	 * 
	 * @param eventID
	 *            The event's ID
	 * @param elapsedTime
	 *            The elapsed time at this event
	 * @param compensationTime
	 *            The compensation time
	 */
	public void eventBoundary(int eventID, int elapsedTime, int compensationTime) {
		if (shouldTransmit(eventID)) {
			internalEventBoundary(eventID, elapsedTime, compensationTime);
		}
	}

	/**
	 * Indicate an event boundary and update counters if necessary
	 * 
	 * @param eventID
	 *            The event's ID
	 */
	public void countingEventBoundary(int eventID) {
		if (collectStats)
			updateEventCounter(eventID);

		if (shouldTransmit(eventID)) {
			if (collectStats)
				elapsedTimer.stop();

			internalEventBoundary(eventID, (int) elapsedTimer.getTime(),
					(int) compensationTimer.getTime());

			if (collectStats) {
				resetTimers();
			}
		}
	}

	/**
	 * Should this event be transmitted?
	 * 
	 * @param eventID
	 *            The event ID
	 * @return true if the server is connected and this event is enabled?
	 */
	public boolean shouldTransmit(int eventID) {
		return isConnected()
				&& getData().getEventFilters().getEnabled()[eventID]
				&& (getData().getEventCount()[eventID] % (getData().getEventFilters().getPeriods()[eventID]) == 0);
	}

	/** ****************** Constructors ******************* */

	/**
	 * Set up a connection. This is called per connection
	 */
	protected void internalSetup() {
		setupEventFilters();

		paused = false;
		pauseNow = false;
		playOne = false;
		shutdownReq = false;
		shutdownSent = false;
	}

	/**
	 * Create a new server interpreter
	 * 
	 * @param name
	 *            The server's name
	 * @param collectStats
	 *            WHether to collect statistics
	 */
	public ServerInterpreter(String name, boolean collectStats, TraceData data) {
		super(name, data);

		cmds[PAUSE_REQ_CMD] = new PauseReqCmd();
		cmds[PAUSE_CMD] = new PauseCmd();
		cmds[RESTART_CMD] = new RestartCmd();
		cmds[PLAY_ONE_CMD] = new PlayOneCmd();
		cmds[SHUTDOWN_REQ_CMD] = new ShutdownReqCmd();
		cmds[SHUTDOWN_CMD] = new ShutdownCmd();
		cmds[STREAM_CMD] = new StreamCmd();
		cmds[EVENT_CMD] = new EventCmd();
		cmds[CONTROL_CMD] = new ControlCmd();
		cmds[EVENT_FILTERS_CMD] = new EventFiltersCmd();
		cmds[EVENT_COUNT_CMD] = new EventCountCmd();
		cmds[SUMMARY_CMD] = new SummaryCmd();
		cmds[SPACE_INFO_CMD] = new SpaceInfoCmd();

		this.collectStats = collectStats;
		if (collectStats) {
			elapsedTimer = new Timer();
			compensationTimer = new Timer();
		}
	}

	/**
	 * Create a new server interpreter
	 * 
	 * @param name
	 *            The server's name
	 * @param collectStats
	 *            Whether to collect statistics
	 * @param events
	 *            The events
	 * @param spaceNum
	 *            The number of spaces
	 */
	public ServerInterpreter(String name, boolean collectStats, TraceData data, Events events,
			int spaceNum) {
		this(name, collectStats, data);
		getData().setSpaceNum(spaceNum);
		getData().setEvents(events);
		setupEventCount();
	}

}
