package gcspy.tools;

import gcspy.interpreter.SingleEventTraceData;
import gcspy.interpreter.Space;
import gcspy.interpreter.Stream;
import gcspy.interpreter.StreamData;

import java.io.IOException;
import java.io.InputStream;


/**
 * Plays a trace from an input stream until the next EVENT_CMD into an object.
 * 
 * @author Thomas Schatzl
 */

public class EventInterpreter extends StreamTrace {

		/*   ****************** Connection ******************* */

		private SingleEventTraceData data;
		private boolean eventFound;
		
		private int lastEventId;
		private int lastElapsedTime;
		private int lastCompensationTime;

		/**
		 * Handle a stream's data and summary values for a stream
		 * 
		 * @param spaceID
		 *            The space's ID
		 * @param streamID
		 *            The stream's ID
		 * @param streamdata
		 *            The stream's data
		 * @param summary
		 *            The summary values
		 * @throws IOException
		 */
		protected void onStreamCmd(int spaceID, int streamID, StreamData streamdata, int summary[])
				throws IOException {
			Stream stream = this.data.getSpace(spaceID).getStream(streamID);
			stream.setData(streamdata);
			stream.setSummary(summary);
		}

		/**
		 * Handle a space's controls
		 * 
		 * @param spaceID
		 *            The space's ID
		 * @param control
		 *            The controls
		 * @throws IOException
		 */
		protected void onControlCmd(int spaceID, byte control[]) throws IOException {
			this.data.getSpace(spaceID).setControl(control);
		}

		/**
		 * Handle event counts
		 * 
		 * @param counts
		 *            The event counts
		 * @throws IOException
		 */
		protected void onEventCountCmd(int counts[]) throws IOException {
			this.data.setEventCount(counts);
		}

		/**
		 * Handle an event
		 * 
		 * @param eventID
		 *            The event's ID
		 * @param elapsedTime
		 *            The elapsed time for the event
		 * @param compensationTime
		 *            Its compensation time
		 * @throws IOException
		 */
		protected void onEventCmd(int eventID, int elapsedTime, int compensationTime)
				throws IOException {
			this.eventFound = true;
			this.lastEventId = eventID;
			this.lastElapsedTime = elapsedTime;
			this.lastCompensationTime = compensationTime;
		}

		/**
		 * Handle space information
		 * 
		 * @param spaceID
		 *            The space's ID
		 * @param spaceInfo
		 *            Its space information
		 * @throws IOException
		 */
		protected void onSpaceInfoCmd(int spaceID, String spaceInfo) throws IOException {
			this.data.getSpace(spaceID).setSpaceInfo(spaceInfo);
		}

		/**
		 * Handle a space
		 * 
		 * @param space
		 *            The space
		 * @throws IOException
		 */
		protected void onSpaceCmd(Space space) throws IOException {
			this.data.setSpace(space.getID(), space);
		}

		public void execute() throws IOException {
			while (!eventFound) {
				parseOne();
			}
		}
		
		public int getLastEventId() {
			return this.lastEventId;
		}
		
		public int getLastElapsedTime() {
			return this.lastElapsedTime;
		}
		
		public int getLastCompensationTime() {
			return this.lastCompensationTime;
		}

		/** ****************** Constructor ******************* */

		public EventInterpreter(InputStream in, SingleEventTraceData data) {
			super(in, StreamTrace.DEFAULT_BUFFER_LEN, StreamTrace.Compression.None);
			this.eventFound = false;
			this.data = data;
		}
}
