/**
 ** $Id: TerminalPlayTrace.java 28 2005-06-20 13:13:35Z rej $
 ** Copyright (c) 2002 Sun Microsystems, Inc.
 **
 ** See the file "license.terms" for information on usage and redistribution
 ** of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 **/

package gcspy.tools;

import gcspy.interpreter.Space;
import gcspy.interpreter.Stream;
import gcspy.interpreter.StreamData;
import gcspy.interpreter.TraceData;
import gcspy.interpreter.server.ServerInterpreter;
import gcspy.utils.Arguments;
import gcspy.utils.Timer;
import gcspy.utils.Utils;

import java.io.FileInputStream;
import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Plays a trace from a file
 * 
 * @author Tony Printezis
 */
public class TerminalPlayTrace extends StreamTrace {
	
	private static final Logger logger = Logger.getLogger(TerminalPlayTrace.class.getName());

	private ServerInterpreter interpreter;
	private TraceData data;
	
	private TraceData getData() {
		return data;
	}

	private boolean preserveTiming = false;
	private int delay = 0;
	private boolean compensate = false;
	private int count;
	private int eventCount[];

	private String filename;

	/*   ****************** Connection ******************* */

	protected void onStreamCmd(int spaceID, int streamID, StreamData data, int summary[])
			throws IOException {
		Stream stream = getData().getSpace(spaceID).getStream(streamID);

		switch (stream.getDataType()) {
		case Byte:
			stream.setData(data);
			break;
		case Short:
			stream.setData(data);
			break;
		case Int:
			stream.setData(data);
			break;
		}
		stream.setSummary(summary);
	}

	protected void onControlCmd(int spaceID, byte control[]) throws IOException {
		getData().getSpace(spaceID).setControl(control);
	}

	protected void onEventCountCmd(int eventCount[]) throws IOException {
		this.eventCount = eventCount;
		getData().setEventCount(eventCount);
	}

	protected void onEventCmd(int eventID, int elapsedTime, int compensationTime) throws IOException {
		if (preserveTiming) {
			// Delay for the time span since the last event occurred
			int millis = elapsedTime - compensationTime;
			if (millis > 0) {
				Utils.sleep(millis);
			}
		} else if (delay > 0) {
			// Delay for a fixed time span
			Utils.sleep(delay);
		}

		++count;

		if (interpreter.shouldTransmit(eventID)) {
			if (preserveTiming && !compensate) {
				// Do not compensate for the time it took to
				// collect and transmit data for this event
				Utils.sleep(compensationTime);
			}

			if (logger.isLoggable(Level.FINE)) {
				logger.log(Level.FINE, count + ". EVENT  '" + getData().getEvents().getName(eventID) + "'  (" + eventID
						+ ")\n" + "    count = " + eventCount[eventID] + ",  elapsed = " + elapsedTime
						+ "ms,  compensation = " + compensationTime + "ms");
			}
		}

		interpreter.eventBoundary(eventID, elapsedTime, compensationTime);
	}

	protected void onSpaceInfoCmd(int spaceID, String spaceInfo) {
		getData().getSpace(spaceID).setSpaceInfo(spaceInfo);
	}

	protected void onSpaceCmd(Space space) throws IOException {
		getData().setSpace(space.getID(), space);
	}
	
	public int getBindPort() {
		return interpreter.getBindPort();
	}
	
	public void bind(int port, boolean wait) throws IOException {
		if (logger.isLoggable(Level.FINE)) {
			String timing;
			if (delay > 0)
				timing = "DELAY " + delay + "ms";
			else if (compensate)
				timing = "COMPENSATED";
			else
				timing = "ORIGINAL";
			
			logger.log(Level.FINE, "Input file: " + filename + ", compression type: " + getCompressionType()
					+ ", timing: " + timing);
		}

		logger.log(Level.FINE, "Reading bootstrap information");
		getData().deserialiseEverything(this);

		logger.log(Level.INFO, "Starting server at port {0}", port);
		interpreter.setSingleServe();
		interpreter.startServer(port, wait);		
	}
	
	public void parseOne() throws IOException {
		super.parseOne();
		interpreter.safepoint();
	}

	public void go() throws IOException {
		Timer timer = new Timer();

		timer.start();
		parse();
		timer.stop();
		
		interpreter.sendShutdown();

		close();

		if (logger.isLoggable(Level.INFO)) {
			logger.log(Level.INFO, "Summary: Read " + count + " events," +
					"playing time was " + (timer.getTime() / 1000) + " secs");
		}
	}

	/** ****************** Constructor ******************* */

	/**
	 * Initialize with the specified trace file path and a fixed delay to impose
	 * between events.
	 */
	public TerminalPlayTrace(String fileName, int delay) throws IOException {
		this(fileName);
		this.delay = delay;
		this.count = 0;
		this.filename = fileName;
	}

	/**
	 * Initialize with the specified trace file path and a parameter that
	 * indicates if processing time of events will be compensated in the replay.
	 * Original delays between events during the recording will be preserved.
	 */
	public TerminalPlayTrace(String fileName, boolean compensate) throws IOException {
		this(fileName);
		this.preserveTiming = true;
		this.compensate = compensate;
		this.filename = fileName;
	}

	/**
	 * Initializes a new {@link TerminalPlayTrace} instance with the specified
	 * trace file path.
	 * 
	 * Events will be replayed without preserving their original delays during
	 * recording.
	 */
	public TerminalPlayTrace(String fileName) throws IOException {
		super(new FileInputStream(fileName), StreamTrace.DEFAULT_BUFFER_LEN, StreamTrace.compressionFor(fileName));

		logger.log(Level.INFO, "Initializing");

		data = new TraceData();
		interpreter = new ServerInterpreter("TerminalPlayTrace", false, data);
	}

	/** ****************** Main ******************* */

	static private void usage(String mesg) {
		System.out.println();
		System.out.println("usage:");
		System.out.println("    gcspy.tools.TerminalPlayTrace <options>");
		System.out.println();
		System.out.println("where <options> are");
		System.out.println("    -port <port>");
		System.out.println("    -input <file name>");
		System.out.println("    -wait (optional)");
		System.out.println("    -delay <delay in ms> (optional)");
		System.out.println("    -compensate (optional)");
		System.out.println("    -verbose (optional)");
		if (mesg != null) {
			reportError(mesg);
		}
		System.exit(-1);
	}

	static private void reportError(String mesg) {
		logger.log(Level.SEVERE, mesg);
	}

	static public void main(String args[]) {
		TerminalPlayTrace playTrace;
		Arguments arguments = new Arguments();
		boolean ignoreDelay = true;
		int delay = 0;
		boolean compensate = false;

		int portType[] = { Arguments.POS_INT_TYPE };
		arguments.add("-port", true, portType);

		int fileType[] = { Arguments.STRING_TYPE };
		arguments.add("-input", true, fileType);

		arguments.add("-wait");

		int delayType[] = { Arguments.POS_INT_TYPE };
		String delayExcluded[] = { "-compensate" };
		arguments.add("-delay", false, delayType, null, delayExcluded);

		String compensateExcluded[] = { "-delay" };
		arguments.add("-compensate", null, compensateExcluded);

		arguments.add("-verbose");

		if (!arguments.parse(args)) {
			usage(arguments.getError());
		}

		Object values[];

		values = arguments.getValues("-port");
		int port = (Integer) values[0];

		values = arguments.getValues("-input");
		String fileName = (String) values[0];

		boolean wait = arguments.isSet("-wait");

		if (arguments.isSet("-delay")) {
			ignoreDelay = false;
			values = arguments.getValues("-delay");
			delay = (Integer) values[0];
		}

		compensate = arguments.isSet("-compensate");

		boolean verbose = arguments.isSet("-verbose");
		if (verbose)
			Logger.getLogger(Logger.GLOBAL_LOGGER_NAME).setLevel(Level.ALL);

		try {
			if (ignoreDelay) {
				playTrace = new TerminalPlayTrace(fileName, compensate);
			} else {
				playTrace = new TerminalPlayTrace(fileName, delay);
			}
			playTrace.bind(port, wait);
			playTrace.go();
		} catch (IOException e) {
			logger.log(Level.SEVERE, "Exception caught, terminating", e);
			System.exit(-1);
		}
	}

}
