/**
 ** $Id: ConnectionDialog.java 28 2005-06-20 13:13:35Z rej $
 ** Copyright (c) 2002 Sun Microsystems, Inc.
 **
 ** See the file "license.terms" for information on usage and redistribution
 ** of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 **/

package gcspy.vis;

import gcspy.utils.FileUtils;
import gcspy.vis.utils.Factory;
import gcspy.vis.utils.OKCancelDialog;

import java.awt.Container;
import java.awt.Dialog;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.net.InetSocketAddress;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextField;

import net.miginfocom.swing.MigLayout;

/**
 * Connection dialog window
 * 
 * @author Tony Printezis
 * @author Thomas Schatzl
 * @author Peter Hofer
 */
public class ConnectionDialog extends OKCancelDialog {

	public static enum ConnectionType {
		SOCKET, FILE,
	}

	private JComboBox typeCB;
	private static final int INDEX_SOCKET = 0;
	private static final int INDEX_FILE = 1;

	private JPanel socketPanel;
	private JTextField hostTF;
	private JTextField portTF;
	private JCheckBox pauseAtStartCB;
	private JCheckBox smallCB;

	private JPanel filePanel;
	private JTextField traceFileTF;
	public static final String DEFAULT_TRACE_FILE = System.getProperty("java.io.tmpdir")
			+ "gcspy-trace.log.gz";
	
	@Override
	protected void onOkClicked() {
		ConnectionType type = getConnectionType();
		if (type == ConnectionType.FILE) {
			if (!FileUtils.exists(getTraceFile())) {
				JOptionPane.showMessageDialog(ConnectionDialog.this,
					"The specified trace file could not be found.");
				return;
			}
		} else if (type == ConnectionType.SOCKET) {
			try {
				int port = getPort();
				if (port == 0) throw new Exception("port = 0");
				InetSocketAddress.createUnresolved(getHost(), port);
			} catch (Exception e) {
				JOptionPane.showMessageDialog(ConnectionDialog.this,
					"The specified remote address is invalid.");
				return;
			}
		}

		super.onOkClicked();
	}

	private ActionListener fileChooseListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			JFileChooser fc = new JFileChooser();
			if (fc.showOpenDialog(ConnectionDialog.this) == JFileChooser.APPROVE_OPTION) {
				traceFileTF.setText(fc.getSelectedFile().getAbsolutePath());
			}
		}
	};

	private ActionListener typeChangedListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			boolean socket = (ConnectionType.SOCKET.equals(getConnectionType()));
			socketPanel.setVisible(socket);
			filePanel.setVisible(!socket);
			pack();
		}
	};

	public String getHost() {
		return hostTF.getText();
	}

	public int getPort() {
		return Integer.parseInt(portTF.getText());
	}

	/**
	 * Should the server pause at the start?
	 * 
	 * @return true if server should pause at the start
	 */
	public boolean getPauseAtStart() {
		return pauseAtStartCB.isSelected();
	}

	/**
	 * Should small tiles be used?
	 * 
	 * @return true is small tiles should be used
	 */
	public boolean getSmall() {
		return smallCB.isSelected();
	}
	
	public String getTraceFile() {
		return traceFileTF.getText();
	}

	public void setConnectionType(ConnectionType type) {
		if (type != getConnectionType()) {
			int index;
			switch (type) {
			case SOCKET:
				index = INDEX_SOCKET;
				break;
			case FILE:
				index = INDEX_FILE;

			default:
				throw new RuntimeException("Unknown connection type");
			}
			typeCB.setSelectedIndex(index); // event handler adjusts view
		}
	}

	public ConnectionType getConnectionType() {
		switch (typeCB.getSelectedIndex()) {
		case INDEX_SOCKET:
			return ConnectionType.SOCKET;
		case INDEX_FILE:
			return ConnectionType.FILE;

		default:
			throw new RuntimeException("Unknown connection type");
		}
	}

	/**
	 * Create a new dialog
	 * 
	 * @param owner
	 *            The parent frame
	 * @param host
	 *            The host to use
	 * @param port
	 *            The port number to use
	 * @param pauseAtStart
	 *            True if the server should pause at the start
	 */
	public ConnectionDialog(Window owner, String host, int port,
			boolean pauseAtStart, boolean autoTrace, String traceFile) {
		
		super(owner);
		
		setTitle("Open Connection");
		setModalityType(Dialog.DEFAULT_MODALITY_TYPE);
		setResizable(false);

		socketPanel = new JPanel(new MigLayout("fill, wrap 2"));
		socketPanel.setBorder(BorderFactory.createTitledBorder("Socket Connection"));
		socketPanel.add(Factory.createLabel("Host:"));
		hostTF = Factory.createTextField(host, 15, true);
		socketPanel.add(hostTF);
		socketPanel.add(Factory.createLabel("Port:"));
		portTF = Factory.createTextField((port != -1) ? "" + port : null, 6,
			true);
		socketPanel.add(portTF);

		filePanel = new JPanel(new MigLayout("fill"));
		filePanel.setBorder(BorderFactory.createTitledBorder("Trace File"));
		filePanel.add(Factory.createLabel("Trace file:"));
		traceFileTF = Factory.createTextField(traceFile, 15, true);
		filePanel.add(traceFileTF);
		JButton button = Factory.createButton("...");
		button.addActionListener(fileChooseListener);
		filePanel.add(button);

		Container container = getContentPane();
		container.setLayout(new MigLayout("fill, hidemode 3, wrap"));

		container.add(Factory.createLabel("Type:"), "split 2");
		typeCB = Factory.createComboBox(typeChangedListener);
		typeCB.addItem("Socket");
		typeCB.addItem("File");
		container.add(typeCB, "gapleft indent, center");

		container.add(socketPanel);
		container.add(filePanel);

		pauseAtStartCB = Factory.createCheckBox("Pause at start", pauseAtStart);
		container.add(pauseAtStartCB);

		container.add(getButtonPanel(), "growx");

		filePanel.setVisible(false);
		pack();
	}
}
