package gcspy.vis;

import gcspy.interpreter.Space;
import gcspy.interpreter.Stream;
import gcspy.utils.Utils;

import java.awt.Color;
import java.util.ArrayList;

/**
 * Generates summary data as {@link PropertyItem} objects for spaces, streams and
 * tiles.
 * 
 * @author Peter Hofer
 */
public class SpaceSummaryGenerator {

	/**
	 * Creates a {@link PropertyItem} object for a stream, typically consisting
	 * of the stream's name and summary value.
	 */
	public static PropertyItem getStreamSummaryItem(Stream stream) {
		int[] summary = stream.getSummary();
		if (summary != null && summary.length > 0) {
			switch (stream.getPresentation()) {
			case Stream.PRESENTATION_PLAIN:
			case Stream.PRESENTATION_PLUS:
			case Stream.PRESENTATION_MAX_VAR:
				return new PropertyItem(stream.getPrefix(), Utils.formatSize(summary[0]) + stream.getSuffix());

			case Stream.PRESENTATION_PERCENT:
			case Stream.PRESENTATION_PERCENT_VAR:
				if (summary[1] == 0) {
					return new PropertyItem(stream.getPrefix(), "N/A %  (" + Utils.formatSize(summary[0])
							+ stream.getSuffix() + ")");
				} else {
					double percentage = summary[0] * 100.0 / summary[1];
					return new PropertyItem(stream.getPrefix(), Utils.formatPercentage(percentage) + "%"
						/* + " (" + Utils.formatSize((long) summary[0])	+ stream.getSuffix() */,
						percentage, null);
				}

			case Stream.PRESENTATION_ENUM:
				return new PropertyItem(stream.getPrefix(), true);
			}
		}

		return null;
	}

	/**
	 * Creates a {@link PropertyItem} object for a single entry of an enumeration
	 * stream.
	 * 
	 * @param index
	 *            Enumeration value.
	 */
	public static PropertyItem getStreamEnumSummaryItem(Stream stream, int index) {
		int[] summary = stream.getSummary();
		if (summary != null && summary.length > 0) {
			String[] enumNames = stream.getEnumNames();

			int summaryTotal = 0;
			for (int i = 0; i < enumNames.length; ++i) {
				summaryTotal += summary[i];
			}

			String name = enumNames[index];
			String value = Integer.toString(summary[index]);
			double percentage = summary[index] * 100.0 / summaryTotal;
			Color color = TileStyle.getEnumerationColor(index);
			return new PropertyItem("   " + name, value, percentage, color);
		}

		return null;
	}

	/**
	 * Creates a summary for a tile consisting of several items.
	 * 
	 * @param space
	 *            The space in which the tile resides.
	 * @param tileIndex
	 *            The index of the tile in the given space.
	 * @return An ordered array of summary items.
	 */
	public static PropertyItem[] getTileSummary(Space space, int tileIndex) {
		ArrayList<PropertyItem> items = new ArrayList<PropertyItem>();

		byte[] control = space.getControl();

		StringBuilder str = new StringBuilder();
		str.append(space.getTitle());
		str.append(tileIndex);
		if (!Space.isControlBackground(control[tileIndex])) {
			str.append(space.getTileName(tileIndex));
			items.add(new PropertyItem(str.toString(), false));

			items.add(new PropertyItem(space.getBlockInfo(), false));
		} else {
			items.add(new PropertyItem(str.toString(), false));
		}

		if (Space.isControlUsed(control[tileIndex])) {
			int nstreams = space.getStreamNum();
			for (int i = 0; i < nstreams; ++i) {
				Stream stream = space.getStream(i);
				if (stream.getAccessor().isDataAvailable()) {
					items.add(getTileStreamSummary(stream, tileIndex));
				}
			}
		} else if (Space.isControlBackground(control[tileIndex])) {
			items.add(new PropertyItem("BACKGROUND", false));
		} else if (Space.isControlUnused(control[tileIndex])) {
			items.add(new PropertyItem(space.getUnusedString(), false));
		}

		PropertyItem[] result = new PropertyItem[items.size()];
		return items.toArray(result);
	}

	/**
	 * Creates a {@link PropertyItem} instance for the value of the given stream
	 * for the given tile.
	 */
	public static PropertyItem getTileStreamSummary(Stream stream, int tileIndex) {
		int value = stream.getAccessor().get(tileIndex);
		String name = stream.getPrefix();
		String suffix = stream.getSuffix();

		switch (stream.getPresentation()) {
		case Stream.PRESENTATION_PLAIN:
		case Stream.PRESENTATION_MAX_VAR:
			return new PropertyItem(name, Utils.formatSize((long) value) + suffix);

		case Stream.PRESENTATION_PLUS: {
			StringBuilder valueStr = new StringBuilder(Utils.formatSize((long) value));
			int max = stream.getMaxValue();
			if (value > max)
				valueStr.append(" (" + Utils.formatSize((long) max) + "+)");
			valueStr.append(suffix);
			return new PropertyItem(name, valueStr.toString());
		}

		case Stream.PRESENTATION_PERCENT: {
			double min = stream.getMinValue();
			double max = stream.getMaxValue();
			double percentage = (value - min) * 100 / (max - min);
			return new PropertyItem(name, Utils.formatSize((long) value) + suffix + "  ("
					+ Utils.formatPercentage(percentage) + "%)", percentage, null);
		}

		case Stream.PRESENTATION_PERCENT_VAR: {
			int max = stream.getMaxValue(tileIndex);
			if (max == 0) {
				return new PropertyItem(name, Utils.formatSize((long) value) + suffix + " (N/A%)");
			} else {
				double percentage = value * 100.0 / max;
				return new PropertyItem(name, Utils.formatSize((long) value) + suffix + " ("
						+ Utils.formatPercentage(percentage) + "%)", percentage, null);
			}
		}

		case Stream.PRESENTATION_ENUM:
			return new PropertyItem(name, stream.getEnumName(value));
		}

		return null;
	}
}
