/**
 ** $Id: TileManagerColors.java 28 2005-06-20 13:13:35Z rej $
 ** Copyright (c) 2002 Sun Microsystems, Inc.
 **
 ** See the file "license.terms" for information on usage and redistribution
 ** of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 **/

package gcspy.vis;

import gcspy.interpreter.Space;
import gcspy.interpreter.Stream;
import gcspy.utils.ColorConstants;

import java.awt.Color;

/**
 * Style configuration for the tile view.
 * 
 * @author Tony Printezis
 * @author Peter Hofer
 */
public class TileStyle {

	public enum EnumPaintStyle {
		COLORED_OPAQUE("Colored Tiles"),
		COLORED_PATTERNS("Colored Patterns"),
		MONOCHROME_PATTERNS("Monochrome Patterns");

		private String description;

		private EnumPaintStyle(String description) {
			this.description = description;
		}

		@Override
		public String toString() {
			return description;
		}
	};
	
	public enum ValuePaintStyle {
		GRADIENT("Fading Colors"),
		VERTICAL_BAR("Vertical Bar");

		private String description;

		private ValuePaintStyle(String description) {
			this.description = description;
		}

		@Override
		public String toString() {
			return description;
		}
	};

	private static Color[] enumColors = null;

	/**
	 * Returns a color to represent a particular value of an enumeration. The
	 * colors provided by this method are intended to be easily distinguishable,
	 * but the sequence will eventually repeat with increasing values (e.g. the
	 * colors returned for 28, 29, 30 might be the same as for 0, 1, 2). Also,
	 * the enumerated values should be zero-based and have no gaps (increments
	 * of 1).
	 * 
	 * @param value
	 *            The enumeration value (may be negative).
	 * @return A color that represents that value.
	 */
	public static Color getEnumerationColor(int value) {
		final int hueSteps = 7; // choose with care
		final float saturation = 0.75f;
		final float brightnessSteps[] = { 1.00f, 0.60f };

		if (enumColors == null) {
			enumColors = new Color[hueSteps * brightnessSteps.length];
		}
		
		int index = (value % enumColors.length);
		if (index < 0) {
			index += enumColors.length;
		}
		
		if (enumColors[index] == null) {
			float h = 1.0f / hueSteps * (index % hueSteps);
			float b = brightnessSteps[index / hueSteps];
			enumColors[index] = Color.getHSBColor(h, saturation, b);
		}
		return enumColors[index];
	}

	/**
	 * Returns a fill pattern to represent a particular value of an enumeration.
	 * 
	 * @param value
	 *            The enumeration value (may be negative).
	 * @return A fill pattern that represents that value.
	 */
	public static FillPattern getEnumerationPattern(int value) {
		final FillPattern[] patterns = FillPattern.ALL_PATTERNS;
		
		int index = (value % patterns.length);
		if (index < 0) {
			index += patterns.length;
		}
		return patterns[index];
	}
	
	private static Color[] enumPatternColors = null;

	/**
	 * Returns a color for display with the fill display for a particular value
	 * of an enumeration.
	 * 
	 * @param value
	 *            The enumeration value (may be negative).
	 * @return A color to display the fill pattern in.
	 */
	public static Color getEnumerationPatternColor(int value) {
		final int hueSteps = 7; // choose with care
		final float saturation = 1.0f;

		if (enumPatternColors == null) {
			enumPatternColors = new Color[hueSteps];
		}

		int index = (value % enumPatternColors.length);
		if (index < 0) {
			index += enumPatternColors.length;
		}

		if (enumPatternColors[index] == null) {
			float h = 1.0f / hueSteps * (index % hueSteps);
			float b = 0.9f;
			enumPatternColors[index] = Color.getHSBColor(h, saturation, b);
		}

		return enumPatternColors[index];
	}

	static private final EnumPaintStyle ENUM_PAINTSTYLE_DEFAULT = EnumPaintStyle.COLORED_OPAQUE;
	static private final ValuePaintStyle VALUE_PAINTSTYLE_DEFAULT = ValuePaintStyle.VERTICAL_BAR;
	static private final Color UNUSED_DEFAULT = Color.LIGHT_GRAY;
	static private final Color SELECTED_DEFAULT = Color.MAGENTA;
	static private final Color MARKER_DEFAULT = Color.ORANGE;
	static private final Color BACKGROUND_DEFAULT = Color.WHITE;
	static private final Color SEPARATOR_DEFAULT = Color.LIGHT_GRAY;
	static private final Color LINK_DEFAULT = Color.GRAY;
	static private final Color STREAM_DEFAULTS[] = {
			Color.RED, ColorConstants.OFF_WHITE,
			Color.MAGENTA, Color.GREEN, Color.ORANGE };

	private Color streamColors[];
	private Color defaultStreamColors[];
	private Color selectedColor;
	private Color markerColor;
	private Color backgroundColor;
	private Color separatorColor;
	private Color linkColor;
	private Color unusedColor;
	private EnumPaintStyle enumPaintStyle;
	private ValuePaintStyle valuePaintStyle;

	public void setSelectedColor(Color selectedColor) {
		this.selectedColor = selectedColor;
	}

	public void setMarkerColor(Color markerColor) {
		this.markerColor = markerColor;
	}

	public void setBackgroundColor(Color backgroundColor) {
		this.backgroundColor = backgroundColor;
	}

	public void setSeparatorColor(Color separatorColor) {
		this.separatorColor = separatorColor;
	}

	public void setLinkColor(Color linkColor) {
		this.linkColor = linkColor;
	}

	/**
	 * Set the color for unused tiles
	 * 
	 * @param unusedColor
	 *            Color of unused tiles
	 */
	public void setUnusedColor(Color unusedColor) {
		this.unusedColor = unusedColor;
	}
	
	public void setEnumPaintStyle(EnumPaintStyle style) {
		this.enumPaintStyle = style;
	}
	
	public void setValuePaintStyle(ValuePaintStyle valueStyle) {
		this.valuePaintStyle = valueStyle;
	}

	/**
	 * Set the color for a stream
	 * 
	 * @param index
	 *            the stream ID
	 * @param streamColor
	 *            the stream color
	 */

	public void setStreamColor(int index, Color streamColor) {
		this.streamColors[index] = streamColor;
	}

	public Color getSelectedColor() {
		return selectedColor;
	}

	public Color getMarkerColor() {
		return markerColor;
	}

	public Color getBackgroundColor() {
		return backgroundColor;
	}

	public Color getSeparatorColor() {
		return separatorColor;
	}

	public Color getLinkColor() {
		return linkColor;
	}

	public Color getUnusedColor() {
		return unusedColor;
	}

	public EnumPaintStyle getEnumPaintStyle() {
		return enumPaintStyle;
	}
	
	public ValuePaintStyle getValuePaintStyle() {
		return valuePaintStyle;
	}

	/**
	 * Return a stream's color
	 * 
	 * @param index
	 *            the stream ID
	 * @return the stream color
	 */
	public Color getStreamColor(int index) {
		return streamColors[index];
	}

	private void setupDefaults(Space space) {
		if (space == null)
			defaultStreamColors = STREAM_DEFAULTS;
		else {
			int len = space.getStreamNum();
			defaultStreamColors = new Color[len];
			for (int i = 0; i < len; ++i) {
				Stream stream = space.getStream(i);
				defaultStreamColors[i] = stream.getColor();
			}
		}
	}

	/**
	 * Revert the colours for tiles, selections, markers, active markers,
	 * separators, links to their defaults
	 */
	public void revertToDefaults() {
		enumPaintStyle = ENUM_PAINTSTYLE_DEFAULT;
		valuePaintStyle = VALUE_PAINTSTYLE_DEFAULT;
		selectedColor = SELECTED_DEFAULT;
		markerColor = MARKER_DEFAULT;
		backgroundColor = BACKGROUND_DEFAULT;
		separatorColor = SEPARATOR_DEFAULT;
		linkColor = LINK_DEFAULT;
		unusedColor = UNUSED_DEFAULT;
		streamColors = new Color[defaultStreamColors.length];
		System.arraycopy(defaultStreamColors, 0, streamColors, 0,
			defaultStreamColors.length);
	}

	/**
	 * Create new set of (default) colors
	 */
	public TileStyle() {
		this(null);
	}

	/**
	 * Create new set of (default) colors for a space
	 * 
	 * @param space
	 *            The space
	 */
	public TileStyle(Space space) {
		setupDefaults(space);
		revertToDefaults();
	}

}
