/**
 ** $Id: TileManagerColorsFrame.java 24 2005-06-17 09:37:56Z rej $
 ** Copyright (c) 2002 Sun Microsystems, Inc.
 **
 ** See the file "license.terms" for information on usage and redistribution
 ** of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 **/

package gcspy.vis;

import gcspy.interpreter.Space;
import gcspy.interpreter.Stream;
import gcspy.vis.TileStyle.EnumPaintStyle;
import gcspy.vis.TileStyle.ValuePaintStyle;
import gcspy.vis.utils.ColorChooser;
import gcspy.vis.utils.ColorChooserListener;
import gcspy.vis.utils.Factory;

import java.awt.Color;
import java.awt.Container;
import java.awt.Font;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;

import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JPanel;

import net.miginfocom.swing.MigLayout;

/**
 * Frame to change the color configuration for the tile manager
 * 
 * @author Tony Printezis
 * @author Peter Hofer
 */
public class TileStyleSettings extends JDialog {

	private ArrayList<TileStyleListener> listeners = new ArrayList<TileStyleListener>();

	private TileStyle styleConfig;

	private boolean updating = false;

	private ColorChooser streamChoosers[];
	private ColorChooser selectedCC;
	private ColorChooser markerCC;
	private ColorChooser backgroundCC;
	private ColorChooser separatorCC;
	private ColorChooser linkCC;
	
	private JComboBox enumStyleCB;
	private JComboBox valueStyleCB;

	private JButton closeB;
	private JButton defaultsB;

	private ActionListener defaultsBListener = new ActionListener() {
		public void actionPerformed(ActionEvent event) {
			styleConfig.revertToDefaults();
			loadConfiguration();
			fireTileColorsChanged();
		}
	};

	private ActionListener closeBListener = new ActionListener() {
		public void actionPerformed(ActionEvent event) {
			setVisible(false);
		}
	};

	private ColorChooserListener chooserListener = new ColorChooserListener() {
		public void colorChosen(ColorChooser source, Color color) {
			if (!updating) {
				updateConfiguration();
				fireTileColorsChanged();
			}
		}
	};

	private ActionListener comboBoxListener = new ActionListener() {
		@Override
		public void actionPerformed(ActionEvent e) {
			if (!updating) {
				updateConfiguration();
				fireTileColorsChanged();
			}
		}
	};

	public void addTileColorsListener(TileStyleListener listener) {
		listeners.add(listener);
	}

	private void fireTileColorsChanged() {
		for (TileStyleListener listener : listeners) {
			listener.tileStyleChanged(this, this.styleConfig);
		}
	}

	private void updateConfiguration() {
		for (int i = 0; i < streamChoosers.length; ++i) {
			if (streamChoosers[i] != null) {
				styleConfig.setStreamColor(i, streamChoosers[i].getColor());
			}
		}
		styleConfig.setSelectedColor(selectedCC.getColor());
		styleConfig.setMarkerColor(markerCC.getColor());
		styleConfig.setBackgroundColor(backgroundCC.getColor());
		styleConfig.setSeparatorColor(separatorCC.getColor());
		styleConfig.setLinkColor(linkCC.getColor());
		styleConfig.setEnumPaintStyle((EnumPaintStyle) enumStyleCB.getSelectedItem());
		styleConfig.setValuePaintStyle((ValuePaintStyle) valueStyleCB.getSelectedItem());
	}

	private void loadConfiguration() {
		updating = true; // prevent indirect recursion in event handlers
		try {
			for (int i = 0; i < streamChoosers.length; ++i) {
				if (streamChoosers[i] != null) {
					streamChoosers[i].setColor(styleConfig.getStreamColor(i));
				}
			}
			selectedCC.setColor(styleConfig.getSelectedColor());
			markerCC.setColor(styleConfig.getMarkerColor());
			backgroundCC.setColor(styleConfig.getBackgroundColor());
			separatorCC.setColor(styleConfig.getSeparatorColor());
			linkCC.setColor(styleConfig.getLinkColor());
			enumStyleCB.setSelectedItem(styleConfig.getEnumPaintStyle());
			valueStyleCB.setSelectedItem(styleConfig.getValuePaintStyle());
		} finally {
			updating = false;
		}
	}

	/**
	 * Create a new frame for choosing tile styles.
	 * 
	 * @param owner
	 *            The parent frame
	 * @param title
	 *            The display title.
	 * @param styleConfig
	 *            Its tile style configuration
	 */
	public TileStyleSettings(Window owner, String title,
			TileStyle styleConfig, Space space, int[] selectedStreams) {
		super(owner, title);

		this.styleConfig = styleConfig;

		Container container = this;
		container.setLayout(new MigLayout("fill", "align center"));

		JLabel titleLabel = Factory.createLabel(space.getFullName(), Font.BOLD);
		container.add(titleLabel, "span");
		
		JPanel streamPanel = new JPanel();
		streamPanel.setLayout(new MigLayout("fill, wrap 2", "[left][right]"));

		int nstreams = space.getStreamNum();
		streamChoosers = new ColorChooser[nstreams];
		JLabel[] nameLabels = new JLabel[nstreams];
		for (int i = 0; i < nstreams; ++i) {
			Stream stream = space.getStream(i);
			
			nameLabels[i] = new JLabel(stream.getName());
			streamPanel.add(nameLabels[i]);
			
			if (stream.getPresentation() != Stream.PRESENTATION_ENUM) {
				ColorChooser chooser = new ColorChooser();
				chooser.addListener(chooserListener);
				streamPanel.add(chooser);
				streamChoosers[i] = chooser;
			} else {
				// Enumeration values are auto-generated
				JLabel label = Factory.createLabel("auto (enumeration)", Font.ITALIC);
				streamPanel.add(label);
			}
		}
		// Highlight selected stream(s)
		for (int stream : selectedStreams) {
			Font font = nameLabels[stream].getFont();
			nameLabels[stream].setFont(font.deriveFont(Font.ITALIC));
		}
		
		container.add(Factory.createTitlePanel("Streams", streamPanel), "growx, span");
		
		JPanel othersPanel = new JPanel();
		othersPanel.setLayout(new MigLayout("fill, wrap 2", "[left][right]"));
		
		othersPanel.add(Factory.createLabel("Background Color:"));
		backgroundCC = new ColorChooser();
		backgroundCC.addListener(chooserListener);
		othersPanel.add(backgroundCC);

		othersPanel.add(Factory.createLabel("Selection Color:"));
		selectedCC = new ColorChooser();
		selectedCC.addListener(chooserListener);
		othersPanel.add(selectedCC);

		othersPanel.add(Factory.createLabel("Marker Color:"));
		markerCC = new ColorChooser();
		markerCC.addListener(chooserListener);
		othersPanel.add(markerCC);

		othersPanel.add(Factory.createLabel("Separator Color:"));
		separatorCC = new ColorChooser();
		separatorCC.addListener(chooserListener);
		othersPanel.add(separatorCC);

		othersPanel.add(Factory.createLabel("Link Color:"));
		linkCC = new ColorChooser();
		linkCC.addListener(chooserListener);
		othersPanel.add(linkCC);

		container.add(Factory.createTitlePanel("Interface Colors", othersPanel), "growx, span");
		
		JPanel stylePanel = new JPanel();
		stylePanel.setLayout(new MigLayout("fill, wrap 2", "[left][right]"));

		stylePanel.add(new JLabel("Enumeration style:"), "push");
		enumStyleCB = new JComboBox(new Object[] {
				TileStyle.EnumPaintStyle.COLORED_OPAQUE,
				TileStyle.EnumPaintStyle.COLORED_PATTERNS,
				TileStyle.EnumPaintStyle.MONOCHROME_PATTERNS });
		enumStyleCB.addActionListener(comboBoxListener);
		stylePanel.add(enumStyleCB, "grow");
		
		stylePanel.add(new JLabel("Value display style:"), "push");
		valueStyleCB = new JComboBox(new Object[] {
				TileStyle.ValuePaintStyle.GRADIENT,
				TileStyle.ValuePaintStyle.VERTICAL_BAR });
		valueStyleCB.addActionListener(comboBoxListener);
		stylePanel.add(valueStyleCB, "grow");

		container.add(stylePanel, "growx, span");

		defaultsB = Factory.createButton("Defaults", true, defaultsBListener);
		container.add(defaultsB, "split 2, sg buttons");

		closeB = Factory.createButton("Close", true, closeBListener);
		container.add(closeB, "sg buttons");

		setResizable(false);
		pack();

		loadConfiguration();
	}

}
